/*
* jacobi2D.c
* 	Hitmap example
* 	Stencil code: Jacobi 2D for the heat equation. Implemented as a 2D cellular automata. 
*
* v2.0
* (c) 2007-2015, Arturo Gonzalez-Escribano
*/

/*
 * <license>
 * 
 * Hitmap v1.2
 * 
 * This software is provided to enhance knowledge and encourage progress in the scientific
 * community. It should be used only for research and educational purposes. Any reproduction
 * or use for commercial purpose, public redistribution, in source or binary forms, with or 
 * without modifications, is NOT ALLOWED without the previous authorization of the copyright 
 * holder. The origin of this software must not be misrepresented; you must not claim that you
 * wrote the original software. If you use this software for any purpose (e.g. publication),
 * a reference to the software package and the authors must be included.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDER AND CONTRIBUTORS "AS IS" AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF 
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL
 * THE AUTHORS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * 
 * Copyright (c) 2007-2015, Trasgo Group, Universidad de Valladolid.
 * All rights reserved.
 * 
 * More information on http://trasgo.infor.uva.es/
 * 
 * </license>
*/

#include <stdio.h>
#include <stdlib.h>
#include <hitmap.h>
#include "CAL.h"
#include "CAL_KernelGPU.h"
#include "timer.h"

hit_tileNewType( float );
HitClock	mainClock;
HitClock	initClock;
HitClock	sequentialClock;
Timer tInner;
Timer TimerStop;
Timer TimerContinue;

/* 9.1 Hit SHAPE EXPAND MANY DIMS IN EQUAL OFFSET */
HitShape hit_shapeExpand(HitShape shape,int dims,int offset){

	// Only works for Signature Shape.
	if(	hit_shapeType(shape) == HIT_CSR_SHAPE ) return HIT_SHAPE_NULL;

	HitShape res = shape;

	int dim;
	for(dim=0;dim<dims;dim++){
		hit_shapeSig(res,dim).begin-=offset;
		hit_shapeSig(res,dim).end+=offset;
	}
	return res;
}

/* 9.1 Hit SHAPE EXPAND A DIM IN A GIVEN DIRECTION */
HitShape hit_shapeDimExpand(HitShape shape,int dim, int position, int offset ){

	// Only works for Signature Shape.
	if(	hit_shapeType(shape) == HIT_CSR_SHAPE ) return HIT_SHAPE_NULL;

	HitShape res = shape;

	if ( position == HIT_SHAPE_BEGIN ) hit_shapeSig(res,dim).begin += offset;
	else hit_shapeSig(res,dim).end += offset;

	return res;
}


/* A. SEQUENTIAL: CELL UPDATE */
//static inline void updateCell( float *myself, float up, float down, float left, float right ) {
//	*myself = ( up + down + left + right ) / 4;
//}

CAL_KERNEL_GPU_CHAR_STATIC(Copy, 2, full, low, low);
CAL_KERNEL_GPU(Copy, 2, OUT, HitTile_float*, dst, IN, HitTile_float*, src){
        int row = threadId.y;
        int col = threadId.x;
        hit_ktileElemAt(dst, 2, row, col) = hit_ktileElemAt(src, 2, row, col);
}




CAL_KERNEL_GPU_CHAR_STATIC(Print, 2, medium, medium, medium);
CAL_KERNEL_GPU(Print, 2, OUT, HitTile_float*, dst, IN, HitTile_float*, src){
        int row = threadId.y ;
        int col = threadId.x ;
	printf("%014.4lf at (%d, %d) \n",hit_ktileElemAt(dst, 2, row, col), row, col);
}


CAL_KERNEL_GPU_CHAR_STATIC(Update, 2, medium, medium, medium);
CAL_KERNEL_GPU(Update, 2, OUT, HitTile_float*, dst, IN, HitTile_float*, src){
        int row = threadId.y + 1;
        int col = threadId.x + 1;
        hit_ktileElemAt(dst, 2, row, col) = (
                        hit_ktileElemAt(src, 2, row - 1, col    ) +
                        hit_ktileElemAt(src, 2, row + 1, col    ) +
                        hit_ktileElemAt(src, 2, row    , col - 1) +
                        hit_ktileElemAt(src, 2, row    , col + 1) ) / 4;
}

	/* 0. OPTIONAL COMPILATION: READING THE INPUT MATRIX FROM A FILE */
#ifdef READ_INPUT
#ifdef READ_BIN
	hit_tileFileRead( &tileMat, "Matrix.in", HIT_FILE_ARRAY );
#else
	hit_tileTextFileRead( &realTileA, "Matrix.in.dtxt", HIT_FILE_ARRAY, HIT_FILE_DOUBLE, 14, 4 );
#endif
#else
	/* 1. INIT Mat = 0 */
	float zero=0; 
	hit_tileFill( &tileMat, &zero );

	/* 2. INIT BORDERS  UP(i)=1, DOWN(i)=2, LEFT(i)=3, RIGHT(i)=4 */
	HitTile root = *hit_tileRoot( &tileMat );
	int i,j;

	/* 2.1. FIRST COLUMN IS MINE */
	if ( hit_sigIn( hit_tileDimSig( tileMat, 1 ), hit_tileDimBegin( root, 1 ) ) )
		hit_tileForDimDomain( tileMat, 0, i )
			hit_tileElemAt( tileMat, 2, i, 0 ) = 3;
		
	/* 2.2. LAST COLUMN IS MINE */
	if ( hit_sigIn( hit_tileDimSig( tileMat, 1 ), hit_tileDimEnd( root, 1 ) ) )
		hit_tileForDimDomain( tileMat, 0, i )
			hit_tileElemAt( tileMat, 2, i, hit_tileDimCard( tileMat, 1 )-1 ) = 4;

	/* 2.3. FIRST ROW IS MINE */
	if ( hit_sigIn( hit_tileDimSig( tileMat, 0 ), hit_tileDimBegin( root, 0 ) ) )
		hit_tileForDimDomain( tileMat, 1, j )
			hit_tileElemAt( tileMat, 2, 0, j ) = 1;
		
	/* 2.4. LAST ROW IS MINE */
	if ( hit_sigIn( hit_tileDimSig( tileMat, 0 ), hit_tileDimEnd( root, 0 ) ) )
		hit_tileForDimDomain( tileMat, 1, j )
			hit_tileElemAt( tileMat, 2, hit_tileDimCard( tileMat, 0 )-1, j ) = 2;
#endif

#ifdef WRITE_INPUT
	/* 4. WRITE MAT ON A FILE */
#ifdef WRITE_BIN
	hit_tileFileWrite( &tileMat, "Matrix.out", HIT_FILE_ARRAY );
#else
#ifdef WRITE_TILE_FILES
	char name[8];
	sprintf(name, "Matrix.%d", hit_Rank );
	hit_tileTextFileWrite( &timeMat, name, HIT_FILE_TILE, HIT_FILE_DOUBLE, 14, 4 );
#endif
	hit_tileTextFileWrite( &tileMat, "Matrix.out.dtxt", HIT_FILE_ARRAY, HIT_FILE_DOUBLE, 14, 4 );
#endif
#endif
}

/* C. MAIN: CELLULAR AUTOMATA */
int main(int argc, char *argv[]) {

	hit_comInit( &argc, &argv );

	int rows, columns, numIter;

	if ( argc != 4 ) {
		fprintf(stderr, "\nUsage: %s <numRows> <numColumns> <numIterations>\n", argv[0]);
		exit(EXIT_FAILURE);
	}
	rows = atoi( argv[1] );
	columns = atoi( argv[2] );
	numIter = atoi( argv[3] );

	/* 0. INIT CLOCKS */
	hit_clockSynchronizeAll();
	hit_clockStart( mainClock );
	hit_clockStart( initClock );
	hit_clockReset( sequentialClock );

    /* 1. CREATE VIRTUAL TOPOLOGY */
    HitTopology topo = hit_topology( plug_topArray2DComplete);


	/* 2. DECLARE FULL MATRIX WITHOUT MEMORY */
	HitTile_float matrix;
	hit_tileDomain( &matrix, float, 2, rows, columns );


    /* 3. COMPUTE PARTITION */
	HitShape parallelShape = hit_tileShape( matrix );
	parallelShape = hit_shapeExpand( parallelShape, 2, -1 );
	HitLayout matLayout = hit_layout( plug_layBlocks, topo, parallelShape );

	/* 4. ACTIVE PROCESSES */
	if ( hit_layImActive( matLayout ) ) {

		/* 4.1. CREATE AND ALLOCATE LOCAL TILES */
		/* 4.1.1. LOCAL TILE WITH SPACE FOR FOREIGN DATA */
		HitTile_float tileMat, tileCopy;
		HitShape expandedShape = hit_shapeDimExpand( hit_layShape(matLayout), 0, HIT_SHAPE_BEGIN, -1 );
		expandedShape = hit_shapeDimExpand( expandedShape, 0, HIT_SHAPE_END, 1 );
		expandedShape = hit_shapeDimExpand( expandedShape, 1, HIT_SHAPE_BEGIN, -1 );
		expandedShape = hit_shapeDimExpand( expandedShape, 1, HIT_SHAPE_END, 1 );
		hit_tileSelect( &tileMat, &matrix, expandedShape );
		hit_tileAlloc( &tileMat );


		/* 4.1.2. COPY OF THE TILE, TO AVOID SEQUENTIAL SEMANTICS ON UPDATES */
		hit_tileSelect( &tileCopy, &tileMat, HIT_SHAPE_WHOLE );
		hit_tileAlloc( &tileCopy );

#ifdef DEBUG
		hit_dumpTileFile( tileMat, "Tile", "Matrix tile" );
#endif

		/* 4.2. COMMUNICATION PATTERN */
		hit_comTagSet( TAG_UP, TAG_DOWN, TAG_LEFT, TAG_RIGHT );
		HitPattern neighSync = hit_pattern( HIT_PAT_UNORDERED );
		hit_patternAdd( &neighSync, 
				hit_comSendRecvSelectTag( matLayout, 
							hit_layNeighbor( matLayout, 0, -1 ) , &tileMat, 
							hit_shape( 2, hit_sigIndex(1), hit_sig( 1, hit_tileDimCard(tileMat,1)-2, 1) ),
							HIT_COM_TILECOORDS, 
							hit_layNeighbor( matLayout, 0, +1 ) , &tileMat, 
							hit_shape( 2, hit_sigIndex( hit_tileDimCard(tileMat,0)-1 ), hit_sig( 1, hit_tileDimCard(tileMat,1)-2, 1) ),
							HIT_COM_TILECOORDS, 
							HIT_DOUBLE, TAG_UP )
							);
		hit_patternAdd( &neighSync, 
				hit_comSendRecvSelectTag( matLayout, 
							hit_layNeighbor( matLayout, 0, +1 ) , &tileMat, 
							hit_shape( 2, hit_sigIndex( hit_tileDimCard(tileMat,0)-2 ), hit_sig( 1, hit_tileDimCard(tileMat,1)-2, 1) ),
							HIT_COM_TILECOORDS, 
							hit_layNeighbor( matLayout, 0, -1 ) , &tileMat, 
							hit_shape( 2, hit_sigIndex(0), hit_sig( 1, hit_tileDimCard(tileMat,1)-2, 1) ),
							HIT_COM_TILECOORDS, 
							HIT_DOUBLE, TAG_DOWN )
							);
		hit_patternAdd( &neighSync, 
				hit_comSendRecvSelectTag( matLayout, 
							hit_layNeighbor( matLayout, 1, -1 ) , &tileMat, 
							hit_shape( 2, hit_sig( 1, hit_tileDimCard(tileMat,0)-2, 1), hit_sigIndex(1) ),
							HIT_COM_TILECOORDS, 
							hit_layNeighbor( matLayout, 1, +1 ) , &tileMat, 
							hit_shape( 2, hit_sig( 1, hit_tileDimCard(tileMat,0)-2, 1), hit_sigIndex( hit_tileDimCard(tileMat,1)-1 ) ),
							HIT_COM_TILECOORDS, 
							HIT_DOUBLE, TAG_LEFT )
							);
		hit_patternAdd( &neighSync, 
				hit_comSendRecvSelectTag( matLayout, 
							hit_layNeighbor( matLayout, 1, +1 ) , &tileMat, 
							hit_shape( 2, hit_sig( 1, hit_tileDimCard(tileMat,0)-2, 1), hit_sigIndex( hit_tileDimCard(tileMat,1)-2 ) ),
							HIT_COM_TILECOORDS, 
							hit_layNeighbor( matLayout, 1, -1 ) , &tileMat, 
							hit_shape( 2, hit_sig( 1, hit_tileDimCard(tileMat,0)-2, 1), hit_sigIndex(0) ),
							HIT_COM_TILECOORDS, 
							HIT_DOUBLE, TAG_RIGHT )
							);

		/* 4.3. INITIALIZE MATRIX (IN PARALLEL) */
		initMatrix( tileMat );
		initMatrix( tileCopy );

		/* 4.3.2 PREPARE CAL */
		CALComm comm;

		//cudaDeviceSynchronize();
	        //TimerStart(tTotal);
	
	        CALThread threads;
	        CALThread(threads, 2, hit_tileDimCard(tileMat,0), hit_tileDimCard(tileMat,1));
	        CALThread update;
	        CALThread(update, 2, hit_tileDimCard(tileMat,0)-2, hit_tileDimCard(tileMat,1)-2);
	//        CALThread(update, 2, 1, 1)
	
	
	        CAL_CommCreate(&comm, CAL_COMM_GPU, 0);
	        //CAL_CommAttach(&comm, (HitTile*)&matrix);
	        //CUDA_CHECK();
	        CAL_CommAttach(&comm, (HitTile*)&tileMat);
	        CUDA_CHECK();
		printf("Falloooooooooooooooooooooooooooooooooooooo \n");
	        CAL_CommInternal(&comm, (HitTile*)&tileCopy);
		printf("Falloooooooooooooooooooooooooooooooooooooo \n");
	        CUDA_CHECK();


	        CAL_CommGPUForceCommToHost( &comm, (HitTile *)&tileMat );

        	HitTile root = *hit_tileRoot( &tileMat );

		/* 4.4. COMPUTATION LOOP */


		int loopIndex;
		for (loopIndex = 0; loopIndex < numIter-1; loopIndex++) {
			// CLOCK: Measure secuential time 
			hit_clockContinue( sequentialClock );

			/* 4.4.1. UPDATE TILE COPY */
			CAL_CommCopyHaloToDevice(&comm, (HitTile *)&tileMat, expandedShape);
                	CUDA_CHECK();
                	cudaDeviceSynchronize();
        	        CAL_CommLaunch(comm, Copy, threads, 2, &tileCopy, &tileMat);
                	CUDA_CHECK();
                	cudaDeviceSynchronize();

			//hit_tileUpdateFromAncestor( &tileCopy );

			/* 4.4.2. SEQUENTIALIZED LOOP */
			CAL_CommLaunch(comm, Update, update, 2, &tileMat, &tileCopy);	 
			CAL_CommCopyBoundaryToHost(&comm, (HitTile *)&tileMat, expandedShape);
			hit_clockStop( sequentialClock );

			/* 4.4.3. COMMUNICATE */
			cudaDeviceSynchronize();
			hit_patternDo( neighSync );
			

		}

		CAL_CommCopyHaloToDevice(&comm, (HitTile *)&tileMat, expandedShape);
               	CUDA_CHECK();
              	cudaDeviceSynchronize();
       	        CAL_CommLaunch(comm, Copy, threads, 2, &tileCopy, &tileMat);
               	CUDA_CHECK();
               	cudaDeviceSynchronize();

		CAL_CommLaunch(comm, Update, update, 2, &tileMat, &tileCopy);	 
		CAL_CommCopyBoundaryToHost(&comm, (HitTile *)&tileMat, expandedShape);


		/* 4.5. LAST ITERATION UPDATE: NO COMMUNICATION AFTER */
		cudaDeviceSynchronize();
		hit_clockContinue( sequentialClock );
		CAL_CommLaunch(comm, Update, update, 2, &tileMat, &tileCopy);	 
		hit_clockStop( sequentialClock );

		/* 4.5.1 CHECK RESULT */
#ifdef Print	
	        int i,j;
                for ( i=0; i<hit_tileDimCard( tileMat, 0 ); i++ )
                        for ( j=0; j<hit_tileDimCard( tileMat, 1 ); j++ )
                                                printf(" %014.4lf at (%d, %d)\n",hit_tileElemAtNoStride( tileMat, 2, i, j ), i, j);
#endif
		/* unroll loop */	
		/* 4.6. CLOCK RESULTS */
		hit_clockStop( mainClock );
		hit_clockReduce( matLayout, mainClock );
		hit_clockReduce( matLayout, sequentialClock );
		hit_clockPrintMax( mainClock );
		hit_clockPrintMax( sequentialClock );
		/* Copy data from controler */
       		CAL_CommDetach(&comm, (HitTile*)&tileMat);
		CUDA_CHECK();
                cudaDeviceSynchronize();
	        CAL_CommDetach(&comm, (HitTile*)&tileCopy);
		CUDA_CHECK();
                cudaDeviceSynchronize();

		/* 4.7. WRITE RESULT MATRIX */
#ifdef WRITE_RESULT
        	root = *hit_tileRoot( &tileMat );
		HitTile_float outputTile;
		HitShape outputShape = hit_tileShape( tileMat );
		/* FIRST ROW IS NOT MINE */
		if ( ! hit_sigIn( hit_shapeSig( outputShape, 0 ), hit_tileDimBegin( root, 0 ) ) )
			outputShape	= hit_shapeDimExpand( outputShape, 0, HIT_SHAPE_BEGIN, +1 );
		/* LAST ROW IS NOT MINE */
		if ( ! hit_sigIn( hit_shapeSig( outputShape, 0 ), hit_tileDimEnd( root, 0 ) ) )
			outputShape = hit_shapeDimExpand( outputShape, 0, HIT_SHAPE_END, -1 );
		/* FIRST COLUMN IS NOT MINE */
		if ( ! hit_sigIn( hit_shapeSig( outputShape, 1 ), hit_tileDimBegin( root, 1 ) ) )
			outputShape = hit_shapeDimExpand( outputShape, 1, HIT_SHAPE_BEGIN, +1 );
		/* LAST COLUMN IS NOT MINE */
		if ( ! hit_sigIn( hit_shapeSig( outputShape, 1 ), hit_tileDimEnd( root, 1 ) ) )
			outputShape = hit_shapeDimExpand( outputShape, 1, HIT_SHAPE_END, -1 );
		hit_tileSelectArrayCoords( &outputTile, &tileMat, outputShape );
#ifdef WRITE_BIN
		hit_tileFileWrite( &outputTile, "Result.out", HIT_FILE_ARRAY );
#else
		hit_tileTextFileWrite( &outputTile, "Result.out.dtxt", HIT_FILE_ARRAY, HIT_FILE_DOUBLE, 14, 4 );
#endif
#endif
		/* 4.8. FREE RESOURCES */
		hit_tileFree( tileMat );
		hit_tileFree( tileCopy );
		hit_patternFree( &neighSync );
	
		hit_layFree( matLayout );
		hit_topFree( topo );
	        CAL_CommDestroy(&comm);

	}
	/* 5. INACTIVE PROCESSES: ONLY COLLECTIVE CLOCK OPERATIONS */
	else {
		hit_clockStop( initClock );
		hit_clockStop( mainClock );
		hit_clockReduce( matLayout, mainClock );
		hit_clockReduce( matLayout, sequentialClock );
		hit_clockPrintMax( mainClock );
		hit_clockPrintMax( sequentialClock );
	}

	/* 6. FREE OTHER RESOURCES */
	hit_comFinalize();
	printf("Finish  \n");
	return 0;
}
