/*
* refMPIJacobi2D.v2.1.c
* 	MPI reference code manually developed and optimized
* 	Stencil code: Jacobi 2D for the heat equation. Implemented as a 2D cellular automata. 
*
*	Allocate only memory needed for the local part 
*	Asynchronous sends and receives
*
* v2.1
* (c) 2007-2015, Arturo Gonzalez-Escribano
*/

/*
 * <license>
 * 
 * Hitmap v1.2
 * 
 * This software is provided to enhance knowledge and encourage progress in the scientific
 * community. It should be used only for research and educational purposes. Any reproduction
 * or use for commercial purpose, public redistribution, in source or binary forms, with or 
 * without modifications, is NOT ALLOWED without the previous authorization of the copyright 
 * holder. The origin of this software must not be misrepresented; you must not claim that you
 * wrote the original software. If you use this software for any purpose (e.g. publication),
 * a reference to the software package and the authors must be included.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDER AND CONTRIBUTORS "AS IS" AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF 
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL
 * THE AUTHORS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * 
 * Copyright (c) 2007-2015, Trasgo Group, Universidad de Valladolid.
 * All rights reserved.
 * 
 * More information on http://trasgo.infor.uva.es/
 * 
 * </license>
*/

#include<stdio.h>
#include<math.h>
#include<stdlib.h>
#include<mpi.h>
#include "timer.h"


#define	mpiTestError(ok,cad)	\
    if ( ok != MPI_SUCCESS ) {	\
        fprintf(stderr,"RunTime-Error, Rank %d: %s - %d\n", mpiRank, cad, ok);\
        exit(-1);				\
    }


int mpiRank;
int mpiNProc;


/* 
 * UTILITY: DECOMPOSE A NUMBER IN 2 PRIME FACTORS AS NEAR AS POSSIBLE TO THE SQUARE ROOT 
 */

__global__ void initMatrixCUDA(double *A, int x, int y)
{

#define elem(mat, idx1, idx2, size)     (mat[(idx1)*size+(idx2)])
        int tx = threadIdx.x;
        int ty = threadIdx.y;

        int bx = blockIdx.x;
        int by = blockIdx.y;

        int bdx = blockDim.x;
        int bdy = blockDim.y;

        int idx = tx + bx * bdx;
        int idy = ty + by * bdy;


        if(( idx < x) && (idy < y)){
                elem( A,idx,idy,y ) = 0.0;
        }
}

__global__ void initHorizontalBorderCUDA(double *A, int x, int y, int pos, int value)
{

#define elem(mat, idx1, idx2, size)     (mat[(idx1)*size+(idx2)])
        int tx = threadIdx.x;
        int ty = threadIdx.y;

        int bx = blockIdx.x;
        int by = blockIdx.y;

        int bdx = blockDim.x;
        int bdy = blockDim.y;


        int gdx = gridDim.x;

        int idg = tx + ty*bdx + (bx + by * gdx) * bdx*bdy;

	if (idg < x-1){
		elem(A, idg, pos, y) = value;
	}
}

__global__ void initVerticalBorderCUDA(double *A, int x, int y, int pos, int value)
{

#define elem(mat, idx1, idx2, size)     (mat[(idx1)*size+(idx2)])
        int tx = threadIdx.x;
        int ty = threadIdx.y;

        int bx = blockIdx.x;
        int by = blockIdx.y;

        int bdx = blockDim.x;
        int bdy = blockDim.y;


        int gdx = gridDim.x;

        int idg = tx + ty*bdx + (bx + by * gdx) * bdx*bdy;

	if (idg < y){
		elem(A, pos, idg, y) = value;
	}
}

__global__ void initColCUDA(double *A, int x)
{

        int tx = threadIdx.x;
        int ty = threadIdx.y;

        int bx = blockIdx.x;
        int by = blockIdx.y;

        int bdx = blockDim.x;
        int bdy = blockDim.y;

        int gdx = gridDim.x;
	
        int idg = tx + ty*bdx + (bx + by * gdx) * bdx*bdy;

        if(idg < x){
                A[idg] = 0.0;
        }
}

__global__ void passColCUDA(double *C, double *Col, int x, int y, int pos)
{

#define elem(mat, idx1, idx2, size)     (mat[(idx1)*size+(idx2)])
        int tx = threadIdx.x;
        int ty = threadIdx.y;

        int bx = blockIdx.x;
        int by = blockIdx.y;

        int bdx = blockDim.x;
        int bdy = blockDim.y;

        int gdx = gridDim.x;
	
        int idg = tx + ty*bdx + (bx + by * gdx) * bdx*bdy;

	if((idg > 0) && (idg < x-1)){
        	elem( C, idg, pos, y) = Col[idg];
        }

}

__global__ void jacobiVerticalBorderCUDA(double *C, double *A, int x, int y, int pos,  double *buffer )
{

#define elem(mat, idx1, idx2, size)     (mat[(idx1)*size+(idx2)])
        int tx = threadIdx.x;
        int ty = threadIdx.y;

	int bx = blockIdx.x;
	int by = blockIdx.y;

	int bdx = blockDim.x;
	int bdy = blockDim.y;

        int gdx = gridDim.x;
	
        int idg = tx + ty*bdx + (bx + by * gdx) * bdx*bdy;

	if((idg > 0) && (idg < x-1)){
       		elem( C, idg, pos, y ) = ( elem( A, idg-1, pos, y )  + elem( A, idg+1, pos, y ) + elem( A, idg, pos-1, y ) + elem( A, idg, pos+1, y ) ) / 4;  
		buffer[idg] = elem(C, idg, pos, y);
	}


}


__global__ void jacobiHorizontalBorderCUDA(double *C, double *A, int x, int y, int pos)
{

#define elem(mat, idx1, idx2, size)     (mat[(idx1)*size+(idx2)])
        int tx = threadIdx.x;
        int ty = threadIdx.y;

	int bx = blockIdx.x;
	int by = blockIdx.y;

	int bdx = blockDim.x;
	int bdy = blockDim.y;

        int gdx = gridDim.x;
	
        int idg = tx + ty*bdx + (bx + by * gdx) * bdx*bdy;

	if((idg > 0)&&(idg < y-1)){
       		elem( C, pos , idg ,y ) = ( elem( A,pos-1, idg ,y )  + elem( A, pos+1, idg,y ) + elem( A,pos, idg-1,y ) + elem( A,pos, idg+1,y ) ) / 4;  
	}

}


__global__ void jacobiCUDA
(double *C, double *A, int x, int y){

#define elem(mat, idx1, idx2, size)     (mat[(idx1)*size+(idx2)])
        int tx = threadIdx.x;
        int ty = threadIdx.y;

	int bx = blockIdx.x;
	int by = blockIdx.y;

	int bdx = blockDim.x;
	int bdy = blockDim.y;

	int idx = tx + bx * bdx;
	int idy = ty + by * bdy;


	if(( idx > 1 ) && ( idy > 1 ) && ( idy < y-2) && (idx < x-2)){
        	elem( C,idx,idy,y ) = ( elem( A,idx-1,idy,y )  + elem( A,idx+1,idy,y ) + elem( A,idx,idy-1,y ) + elem( A,idx,idy+1,y ) ) / 4;  
	}

}



int * factors2D(int numProc) {
	int *result;
	int low, high;
	int product;

	/* 1. ALLOCATE MEMORY FOR result */
	result = (int *)calloc(2, sizeof(int));
	if ( result == NULL )
        fprintf(stderr,"RunTime-Error, Rank %d: Allocating factors2D result\n", mpiRank);

	/* 2. START AS NEAR AS POSSIBLE TO THE SQUARE ROOT */
	low = high = (int)sqrt((double)numProc);
	product = low*high;

	/* 3. MOVE UP/DOWN TO SEARCH THE NEAREST FACTORS */
	while ( product != numProc ) {
		if (product < numProc) {
			high++;
			product = product + low;
		}
		else {
			low--;
			if (low==1) { high=numProc; product=numProc; }
			else product = product - high;
		}
	}

	/* 4. RETURN RESULT */
	result[0] = high;
	result[1] = low;
	return result;
}


/*
 * MAIN PROGRAM
 */
int main(int argc, char *argv[]) {
	int ok;
	int i,j;
#ifdef DEBUG
	setbuf(stderr, NULL);
	setbuf(stdout, NULL);
#endif

	/* 0. READ PARAMETERS */
	int rows, columns, numIter;

	if ( argc != 4 ) {
		fprintf(stderr, "\nUsage: %s <numRows> <numColumns> <numIterations>\n", argv[0]);
		exit(EXIT_FAILURE);
	}
	rows = atoi( argv[1] );
	columns = atoi( argv[2] );
	numIter = atoi( argv[3] );


    /* 1. INITIALIZE MPI */
    ok = MPI_Init( &argc, &argv);
	mpiTestError( ok, "Initializing MPI" );

    /* 2. GET BASIC GROUP PARAMETERS */
    MPI_Comm_rank(MPI_COMM_WORLD, &mpiRank);
    MPI_Comm_size(MPI_COMM_WORLD, &mpiNProc);

        /* 2.1 SET CUDA PROPERTIES */

	int nGPU; 
	cudaGetDeviceCount(&nGPU);
        int devID = (mpiRank+1)%nGPU;

	cudaSetDevice(devID);

	cudaError_t error;
        cudaDeviceProp deviceProp;
        cudaGetDeviceProperties(&deviceProp, devID);


	cudaStream_t stream0, stream1, stream2, stream3, stream4;
	cudaStreamCreate( &stream0 );
	cudaStreamCreate( &stream1 );
	cudaStreamCreate( &stream2 );
	cudaStreamCreate( &stream3 );
	cudaStreamCreate( &stream4 );


	/* 3. START CLOCK */
	MPI_Barrier( MPI_COMM_WORLD );
	/* Timer */
	Timer tLoop;
	
	cudaEvent_t start, stop;
	cudaEventCreate(&start);
	cudaEventCreate(&stop);
	float milliseconds = 0;
	float tCompTotal = 0;

	/* 4. CARTESIAN TOPOLOGY */
	MPI_Comm cartesianComm;
	int *procs = factors2D( mpiNProc );

	/* 4.1. CHECK: NO SUPPORT FOR PARTITIONING LESS ELEMENTS THAN PROCS IN A DIMENSION */
	if ( procs[0] > rows-2 || procs[1] > columns-2 ) {
		if ( mpiRank == 0 ) fprintf(stderr, "Error: This code does not support partitioning less rows/columns than processors in that axis\n");
		MPI_Finalize();
	}

	int periods[2] = { 0, 0 };
	ok = MPI_Cart_create(MPI_COMM_WORLD, 2, procs, periods, 0, &cartesianComm);
	mpiTestError( ok, "Creating cartesian toology" );

	int myCoords[2];
	ok = MPI_Cart_coords(cartesianComm, mpiRank, 2, myCoords);
	mpiTestError( ok, "Getting cartesian coordinates" );

	int rankUp=-1000, rankDown=-1000, rankLeft=-1000, rankRight=-1000;
	int foo;
	if ( myCoords[0] != 0 ) {
		ok = MPI_Cart_shift(cartesianComm, 0, -1, &foo, &rankUp);
		mpiTestError( ok, "Shifting cartesian coordinates up" );
	}

	if ( myCoords[0] != procs[0]-1 ) {
		ok = MPI_Cart_shift(cartesianComm, 0, +1, &foo, &rankDown);
		mpiTestError( ok, "Shifting cartesian coordinates down" );
	}

	if ( myCoords[1] != 0 ) {
		ok = MPI_Cart_shift(cartesianComm, 1, -1, &foo, &rankLeft);
		mpiTestError( ok, "Shifting cartesian coordinates left" );
	}

	if ( myCoords[1] != procs[1]-1 ) {
		ok = MPI_Cart_shift(cartesianComm, 1, +1, &foo, &rankRight);
		mpiTestError( ok, "Shifting cartesian coordinates right" );
	}

#ifdef DEBUG
fprintf(stderr,"CTRL RANKS: %d (%d,%d,%d,%d)\n", mpiRank, rankUp, rankDown, rankLeft, rankRight);
#endif

	/* 5. COMPUTE THE LOCAL PART OF THE MATRIX */
	double procRatio[2];
	procRatio[0] = (double)(rows-2)/procs[0];
	procRatio[1] = (double)(columns-2)/procs[1];

	int begin[2];
	int end[2];

#ifdef DEBUG
fprintf(stderr,"CTRL %d, myCoords %d,%d\n", mpiRank, myCoords[0], myCoords[1]);
#endif

	begin[0] = 1 + (int)floor( myCoords[0]*procRatio[0] ); 
	end[0] = 1 + (int)floor( (myCoords[0]+((rows > procs[0]) ? 1 : 0))*procRatio[0] - ((rows > procs[0]) ? 1 : 0) );
	begin[1] = 1 + (int)floor( myCoords[1]*procRatio[1] ); 
	end[1] = 1 + (int)floor( (myCoords[1]+((columns > procs[1]) ? 1 : 0))*procRatio[1] - ((columns > procs[1]) ? 1 : 0) );

#ifdef DEBUG
fprintf(stderr,"CTRL %d, matrix[%d:%d][%d:%d]\n", mpiRank, begin[0], end[0], begin[1], end[1]);
#endif

    /* 6. DEFINE MAIN DATA STRUCTURES */
	int size[2] = { end[0]-begin[0]+1+2, end[1]-begin[1]+1+2 };
	double *matrix;
        double *matrixCopy;
        double *leftCol, *leftColRecv;
        double *rightCol, *rightColRecv;
        double *d_matrix, *d_matrixCopy;
        double *d_leftCol, *d_rightCol;
        double *d_leftColRecv, *d_rightColRecv;

        unsigned int mem_size_matrix = (size_t)(size[0] * size[1])* sizeof(double);
	unsigned int mem_size_col = (size_t)size[0]*sizeof(double);
	unsigned int mem_size_border_updown = (size_t)(size[1]-2)*sizeof(double);

        cudaHostAlloc((void**) &leftCol, mem_size_col, cudaHostAllocDefault);
        cudaHostAlloc((void**) &rightCol, mem_size_col, cudaHostAllocDefault);
        cudaHostAlloc((void**) &leftColRecv, mem_size_col, cudaHostAllocDefault);
        cudaHostAlloc((void**) &rightColRecv, mem_size_col, cudaHostAllocDefault);
        cudaHostAlloc((void**) &matrix, mem_size_matrix, cudaHostAllocDefault);
        cudaHostAlloc((void**) &matrixCopy, mem_size_matrix, cudaHostAllocDefault);
        cudaMalloc((void **) &d_matrix, mem_size_matrix);
        cudaMalloc((void **) &d_matrixCopy, mem_size_matrix);



        cudaMalloc((void **) &d_leftCol, mem_size_col);
        cudaMalloc((void **) &d_rightCol, mem_size_col);
        cudaMalloc((void **) &d_leftColRecv, mem_size_col);
        cudaMalloc((void **) &d_rightColRecv, mem_size_col);
        int block_size = (deviceProp.major < 2) ? 16 : 32;

	dim3 threadsCol(block_size, block_size);
	dim3 gridCol((size[0]/threadsCol.x)+1, 1);
        dim3 threads(block_size, block_size);
        dim3 grid((size[0] / threads.x)+1 , (size[1] / threads.y)+1 );




	if ( matrix==NULL || matrixCopy==NULL) {
        fprintf(stderr,"RunTime-Error, Rank %d: Allocating data structures\n", mpiRank);
		exit(2);
	}

#define	elementGlobal(mat, idx1, idx2)	(mat[(idx1)*columns+(idx2)])
#define	element(mat, idx1, idx2)	(mat[(idx1)*size[1]+(idx2)])

	/* 7. INITIAL MATRIX VALUES */

        initMatrixCUDA<<< grid, threads>>>( d_matrix, size[0], size[1]);
        initMatrixCUDA<<< grid, threads>>>( d_matrixCopy, size[0], size[1]);
	initColCUDA<<< grid, threads>>>(d_leftCol, size[0]);
	initColCUDA<<< grid, threads>>>(d_leftColRecv, size[0]);
	initColCUDA<<< grid, threads>>>(d_rightCol, size[0]);
	initColCUDA<<< grid, threads>>>(d_rightColRecv, size[0]);
	

	/* 8. FIXED BOUNDARY VALUES */
	if ( myCoords[1] == 0 ) {
		initHorizontalBorderCUDA<<< gridCol, threadsCol>>>( d_matrix, size[0], size[1], 0, 3.0);
		initHorizontalBorderCUDA<<< gridCol, threadsCol>>>( d_matrixCopy, size[0], size[1], 0, 3.0);
	}
	if ( myCoords[1] == procs[1]-1 ) {
		initHorizontalBorderCUDA<<< gridCol, threadsCol>>>( d_matrix, size[0], size[1], size[1]-1, 4.0);
		initHorizontalBorderCUDA<<< gridCol, threadsCol>>>( d_matrixCopy, size[0], size[1], size[1]-1, 4.0);
	}
	if ( myCoords[0] == 0 ) {
		initVerticalBorderCUDA<<< gridCol, threadsCol >>>(d_matrix, size[0], size[1], 0, 1.0);
		initVerticalBorderCUDA<<< gridCol, threadsCol >>>(d_matrixCopy, size[0], size[1], 0, 1.0);
	}
	if ( myCoords[0] == procs[0]-1 ) {
		initVerticalBorderCUDA<<< gridCol, threadsCol >>>(d_matrix, size[0], size[1], size[0]-1, 2.0);
		initVerticalBorderCUDA<<< gridCol, threadsCol >>>(d_matrixCopy, size[0], size[1], size[0]-1, 2.0);
	}

	/* 9. COMMIT MPI TYPES */
	MPI_Datatype verticalBorderType;
        ok = MPI_Type_hvector( size[0]-2, 1, (int)sizeof(double)*size[1], MPI_DOUBLE, &verticalBorderType );
        mpiTestError( ok, "Creating the columns type" );
        ok = MPI_Type_commit( &verticalBorderType );
        mpiTestError( ok, "Commiting the columns type" );


	MPI_Request request[8] = { MPI_REQUEST_NULL, MPI_REQUEST_NULL, MPI_REQUEST_NULL, MPI_REQUEST_NULL, MPI_REQUEST_NULL, MPI_REQUEST_NULL, MPI_REQUEST_NULL, MPI_REQUEST_NULL };
	MPI_Status statusV[8];
/*
        if ( mpiRank == 0 ) {
                FILE *f;
                if ((f=fopen("Whats Happening","w")) == NULL) {
                        fprintf(stderr,"Error: Imposible to open output file\n");
                        exit(-1);
                }
                fprintf(f,"help" );
                fclose(f);
        }
*/
	/* 10. LOOP: UNROLLED ONE ITERATION TO AVOID EXTRA COMM. STAGE */
	int loopIndex;
	TimerStart( tLoop );
	TimerStop( tLoop );

	for (loopIndex = 0; loopIndex < numIter-1; loopIndex++) {

		TimerContinue( tLoop );
                jacobiCUDA<<< grid, threads, 0, stream0>>>( d_matrix, d_matrixCopy, size[0], size[1]);
                jacobiHorizontalBorderCUDA<<< gridCol, threadsCol, 0, stream1>>>( d_matrix, d_matrixCopy, size[0], size[1], 1);
                jacobiHorizontalBorderCUDA<<< gridCol, threadsCol, 0, stream2>>>( d_matrix, d_matrixCopy, size[0], size[1], size[0]-2);
                jacobiVerticalBorderCUDA<<< gridCol, threadsCol, 0, stream3>>>( d_matrix, d_matrixCopy, size[0], size[1], 1, d_leftCol);
                jacobiVerticalBorderCUDA<<< gridCol, threadsCol, 0, stream4>>>( d_matrix, d_matrixCopy, size[0], size[1], size[1]-2, d_rightCol);

		/* 12. COMMUNICATE BORDERS: SEND MATRIX VALUES */
		
		/* Copy from device to host */
		if ( myCoords[0] != 0 ) {

			error = cudaMemcpyAsync( &element(matrix ,1, 1), &element(d_matrix, 1, 1), mem_size_border_updown, cudaMemcpyDeviceToHost, stream1);
			if (error != cudaSuccess)
			{
        			printf("cudaMemcpy (h_C,d_C0) returned error %s (code %d), line(%d)\n", cudaGetErrorString(error), error, __LINE__);
        			exit(EXIT_FAILURE);
			}
		}
		if ( myCoords[0] != procs[0]-1 ) {
			error = cudaMemcpyAsync( &element(matrix ,size[0]-2, 1), &element(d_matrix, size[0]-2, 1), mem_size_border_updown, cudaMemcpyDeviceToHost, stream2);
			if (error != cudaSuccess)
			{
        			printf("cudaMemcpy (h_C,d_C0) returned error %s (code %d), line(%d)\n", cudaGetErrorString(error), error, __LINE__);
        			exit(EXIT_FAILURE);
			}
		}
		if ( myCoords[1] != 0 ) {
			error = cudaMemcpyAsync(leftCol, d_leftCol, mem_size_col, cudaMemcpyDeviceToHost, stream3);
			if (error != cudaSuccess)
			{
        			printf("cudaMemcpy (h_C,d_C0) returned error %s (code %d), line(%d)\n", cudaGetErrorString(error), error, __LINE__);
        			exit(EXIT_FAILURE);
			}	
		}
		if ( myCoords[1] != procs[1]-1 ) {
			error = cudaMemcpyAsync(rightCol, d_rightCol, mem_size_col, cudaMemcpyDeviceToHost, stream4);
			if (error != cudaSuccess)
			{
        			printf("cudaMemcpy (h_C,d_C0) returned error %s (code %d), line(%d)\n", cudaGetErrorString(error), error, __LINE__);
        			exit(EXIT_FAILURE);
			}
		}

		/* Send data to neighbours */
		cudaStreamSynchronize(stream1);
		cudaStreamSynchronize(stream2);
		cudaStreamSynchronize(stream3);
		cudaStreamSynchronize(stream4);

		if ( myCoords[0] != 0 ) {
			ok = MPI_Isend( &element(matrix, 1, 1), size[1]-2, MPI_DOUBLE, rankUp, 0, MPI_COMM_WORLD, &request[0] );
			mpiTestError( ok, "Send Up" );
		}
		if ( myCoords[0] != procs[0]-1 ) {
			ok = MPI_Isend( &element(matrix, size[0]-2, 1), size[1]-2, MPI_DOUBLE, rankDown, 1, MPI_COMM_WORLD, &request[1] );
			mpiTestError( ok, "Send Down" );
		}
		if ( myCoords[1] != 0 ) {
			ok = MPI_Isend( leftCol, size[0], MPI_DOUBLE, rankLeft, 2, MPI_COMM_WORLD, &request[2] );
			mpiTestError( ok, "Send Left" );
		}
		if ( myCoords[1] != procs[1]-1 ) {
			ok = MPI_Isend( rightCol, size[0], MPI_DOUBLE, rankRight, 3, MPI_COMM_WORLD, &request[3] );
			mpiTestError( ok, "Send Right" );
		}

		/* 14. COMMUNICATE BORDERS: RECV IN MATRIX COPY */
		if ( myCoords[0] != procs[0]-1 ) {
			ok = MPI_Irecv( &element(matrix, size[0]-1, 1), size[1]-2, MPI_DOUBLE, rankDown, 0, MPI_COMM_WORLD, &request[4] );
			mpiTestError( ok, "Recv from Down" );
		}
		if ( myCoords[0] != 0 ) {
			ok = MPI_Irecv( &element(matrix, 0, 1), size[1]-2, MPI_DOUBLE, rankUp, 1, MPI_COMM_WORLD, &request[5] );
			mpiTestError( ok, "Recv from Up" );
		}
		if ( myCoords[1] != procs[1]-1 ) {
			ok = MPI_Irecv( rightColRecv, size[0], MPI_DOUBLE, rankRight, 2, MPI_COMM_WORLD, &request[6] );
			mpiTestError( ok, "Recv from Right" );
		}
		if ( myCoords[1] != 0 ) {
			ok = MPI_Irecv( leftColRecv, size[0], MPI_DOUBLE, rankLeft, 3, MPI_COMM_WORLD, &request[7] );
			mpiTestError( ok, "Recv from Left" );
		}

		/* Wait to receive all data */
		

		MPI_Waitall( 8, request, statusV );

		if ( myCoords[0] != procs[0]-1 ) {
			error = cudaMemcpyAsync( &element(d_matrix, size[0]-1, 1), &element(matrix, size[0]-1, 1), mem_size_border_updown, cudaMemcpyHostToDevice, stream1);
			if (error != cudaSuccess)
			{
				printf("%d %d %d \n", sizeof(matrix), sizeof(d_matrix), mem_size_border_updown);
        			printf("cudaMemcpy (h_C,d_C0) returned error %s (code %d), line(%d)\n", cudaGetErrorString(error), error, __LINE__);
        			exit(EXIT_FAILURE);
			}
		}
		if ( myCoords[0] != 0 ) {
			error = cudaMemcpyAsync( &element(d_matrix, 0, 1), &element(matrix, 0, 1), mem_size_border_updown, cudaMemcpyHostToDevice, stream2);
			if (error != cudaSuccess)
			{
        			printf("cudaMemcpy (h_C,d_C0) returned error %s (code %d), line(%d)\n", cudaGetErrorString(error), error, __LINE__);
        			exit(EXIT_FAILURE);
			}
		}
		if ( myCoords[1] != procs[1]-1 ) {
			error = cudaMemcpyAsync(d_rightColRecv, rightColRecv, mem_size_col, cudaMemcpyHostToDevice, stream3);
			if (error != cudaSuccess)
			{
        			printf("cudaMemcpy (h_C,d_C0) returned error %s (code %d), line(%d)\n", cudaGetErrorString(error), error, __LINE__);
        			exit(EXIT_FAILURE);
			}
			passColCUDA<<< grid, threads, 0, stream1>>>( d_matrix, d_rightColRecv, size[0], size[1], size[1]-1);
		}
		if ( myCoords[1] != 0 ) {
			error = cudaMemcpyAsync(d_leftColRecv, leftColRecv, mem_size_col, cudaMemcpyHostToDevice, stream4);
			if (error != cudaSuccess)
			{
        			printf("cudaMemcpy (h_C,d_C0) returned error %s (code %d), line(%d)\n", cudaGetErrorString(error), error, __LINE__);
        			exit(EXIT_FAILURE);
			}
                	passColCUDA<<< grid, threads, 0, stream1>>>( d_matrix, d_leftColRecv, size[0], size[1], 0);
		}

	
	
		/* 15. END Communications */
		cudaDeviceSynchronize();
		//MPI_Waitall( 8, request, statusV );

		TimerStop( tLoop );
		cudaEventElapsedTime(&milliseconds, start, stop);
		tCompTotal = tCompTotal + milliseconds;
		/* Swap pointers */ 

		double* aux = matrix;
		matrix = matrixCopy;
		matrixCopy = aux;

		double* d_aux = d_matrix;
		d_matrix = d_matrixCopy;
		d_matrixCopy = d_aux;


	} /* END LOOP */

	TimerStop( tLoop );	
	// UNROLLED LOOP: LAST ITERATION WITHOUT COMMUNICATION
	/* 16. COMPUTE LOCAL PART */
	if(loopIndex%2==0){	
		double* aux = matrix;
		matrix = matrixCopy;
		matrixCopy = aux;

		double* d_aux = d_matrix;
		d_matrix = d_matrixCopy;
		d_matrixCopy = d_aux;
	}

       	//cudaMemcpy(matrix, d_matrix, mem_size_matrix, cudaMemcpyDeviceToHost);
       	cudaMemcpy(matrixCopy, d_matrixCopy, mem_size_matrix, cudaMemcpyDeviceToHost);
       	cudaMemcpy(matrix, d_matrix, mem_size_matrix, cudaMemcpyDeviceToHost);
#ifdef DEBUG
for (i=1; i<size[0]-1; i++) {
	for (j=1; j<size[1]-1; j++) {
		fprintf(stderr,"M[%d][%d] %lf\n", i,j, element(matrix,i,j));
	}
}
#endif

	double *matrixResult = NULL;

	/* 18. COMMUNICATE ALL LOCAL PIECES */
	if ( mpiRank == 0 ) {
		matrixResult = (double *)calloc( (size_t)(rows * columns), sizeof(double) );

		/* 17.0. COPY MY OWN RESULTS */
		for (i=0; i<size[0]; i++) {
			for (j=0; j<size[1]; j++) {
				elementGlobal( matrixResult,i,j ) = element( matrix,i,j );
			}
		}

		/* 18.1. RANK 0 RECEIVES ALL */
		int rank;
		for ( rank=1; rank<mpiNProc; rank++ ) {
                        /* COMPUTE THE POSITION OF THE REMOTE LOCAL PART TO BE RECEIVED */
                        int remoteCoords[2];
                        ok = MPI_Cart_coords(cartesianComm, rank, 2, remoteCoords);
                        mpiTestError( ok, "Getting remote cartesian coordinates in end comm." );

                        int remoteBegin[2];
                        int remoteEnd[2];

                        remoteBegin[0] = 1 + (int)floor( remoteCoords[0]*procRatio[0] ); 
                        remoteEnd[0] = 1 + (int)floor( (remoteCoords[0]+((rows > procs[0]) ? 1 : 0))*procRatio[0] - ((rows > procs[0]) ? 1 : 0) );
                        remoteBegin[1] = 1 + (int)floor( remoteCoords[1]*procRatio[1] ); 
                        remoteEnd[1] = 1 + (int)floor( (remoteCoords[1]+((columns > procs[1]) ? 1 : 0))*procRatio[1] - ((columns > procs[1]) ? 1 : 0) );

                        // INCLUDE BORDERS
                        if ( remoteBegin[0] == 1 ) remoteBegin[0]--;
                        if ( remoteBegin[1] == 1 ) remoteBegin[1]--;
                        if ( remoteEnd[0] == rows-2 ) remoteEnd[0]++;
                        if ( remoteEnd[1] == columns-2 ) remoteEnd[1]++;

#ifdef DEBUG
fprintf(stderr,"CTRL RemoteMatrix %d, [%d:%d][%d:%d]\n", rank, remoteBegin[0], remoteEnd[0], remoteBegin[1], remoteEnd[1]);
#endif
                        MPI_Status status;
                        MPI_Datatype remoteType;
                        ok = MPI_Type_hvector( remoteEnd[0]-remoteBegin[0]+1, remoteEnd[1]-remoteBegin[1]+1, (MPI_Aint)((int)sizeof(double)*columns), MPI_DOUBLE, &remoteType );
                        mpiTestError( ok, "Getting remote type in end comm." );
                        ok = MPI_Type_commit( &remoteType );
                        mpiTestError( ok, "Commiting the remote type in end comm." );

                        ok = MPI_Recv( &elementGlobal( matrixResult,remoteBegin[0],remoteBegin[1] ), 1, remoteType, rank, rank, MPI_COMM_WORLD, &status );
                        mpiTestError( ok, "Receiving remote parts in end comm." );

                        ok = MPI_Type_free( &remoteType );
                        mpiTestError( ok, "Freeing remote type in end comm." );
                }

	}
	else {
		/* 18.2. OTHERS SEND THE LOCAL PART */
#ifdef DEBUG
fprintf(stderr,"CTRL LocalMatrix %d, [%d:%d][%d:%d]\n", mpiRank, begin[0], end[0], begin[1], end[1]);
#endif

		// INCLUDE BORDERS
                int localBegin[2];
                int localEnd[2];
                localBegin[0] = ( begin[0] == 1 ) ? 0 : 1;
                localBegin[1] = ( begin[1] == 1 ) ? 0 : 1;
                localEnd[0] = ( end[0] == rows-2 ) ? size[0]-1 : size[0]-2;
                localEnd[1] = ( end[1] == columns-2 ) ? size[1]-1 : size[1]-2;

                MPI_Datatype localType;
                ok = MPI_Type_hvector( localEnd[0]-localBegin[0]+1, localEnd[1]-localBegin[1]+1, (int)sizeof(double)*size[1], MPI_DOUBLE, &localType );
                mpiTestError( ok, "Getting local type in end comm." );
                ok = MPI_Type_commit( &localType );
                mpiTestError( ok, "Commiting the local type in end comm." );

                ok = MPI_Send( &element( matrix,localBegin[0],localBegin[1] ), 1, localType, 0, mpiRank, MPI_COMM_WORLD );
                mpiTestError( ok, "Sending local part in end comm." );

                ok = MPI_Type_free( &localType );
                mpiTestError( ok, "Freeing local type in end comm." );
        }

	/* 19. STOP AND PRINT CLOCK */	
	if ( mpiRank == 0)
		printf("%.8lf\n", TimerGetTime(tLoop) );
		//printf("%.8lf\n", TimerGetTime(tLoop) );
	/* 20. WRITE RESULT IN FILE */
#ifdef DEBUG
	if ( mpiRank == 0 ) {
		FILE *f;
		if ((f=fopen("Result.out.dtxt","w")) == NULL) {
			fprintf(stderr,"Error: Imposible to open output file\n");
			exit(-1);
		}
		for (i=0; i<rows; i++) {
			for (j=0; j<columns; j++) {
				fprintf(f,"%014.4lf", elementGlobal(matrixResult,i,j) );
				fprintf(f,"\n");
			}
		}
		fclose(f);
	}
#endif

    /* 21. END PROGRAM */
	free( procs );
	cudaFreeHost( matrix );
	cudaFreeHost( matrixCopy );
	cudaFreeHost( leftCol );
	cudaFreeHost( leftColRecv );
	cudaFreeHost( rightCol );
	cudaFreeHost( rightColRecv );
	cudaFree( d_matrix );
	cudaFree( d_matrixCopy );
	cudaFree( d_leftCol );
	cudaFree( d_leftColRecv );
	cudaFree( d_rightCol );
	cudaFree( d_rightColRecv );
	if ( mpiRank == 0 ) free( matrixResult );
	MPI_Comm_free( &cartesianComm );
	MPI_Finalize();

	/* 22. RETURN CORRECT */
	return 0;
}
