/**
 * Copyright 1993-2015 NVIDIA Corporation.  All rights reserved.
 *
 * Please refer to the NVIDIA end user license agreement (EULA) associated
 * with this source code for terms and conditions that govern your use of
 * this software. Any use, reproduction, disclosure, or distribution of
 * this software and related documentation outside the terms of the EULA
 * is strictly prohibited.
 *
 */

/**
 * Matrix multiplication: C = A * B.
 * Host code.
 *
 * This sample implements matrix multiplication as described in Chapter 3
 * of the programming guide.
 * It has been written for clarity of exposition to illustrate various CUDA
 * programming principles, not with the goal of providing the most
 * performant generic kernel for matrix multiplication.
 *
 * See also:
 * V. Volkov and J. Demmel, "Benchmarking GPUs to tune dense linear algebra,"
 * in Proc. 2008 ACM/IEEE Conf. on Supercomputing (SC '08),
 * Piscataway, NJ: IEEE Press, 2008, pp. Art. 31:1-11.
 */

// System includes
#include <stdio.h>
#include <assert.h>
#include "timer.h"

#define nMatrix 500

template <int BLOCK_SIZE> __global__ void
matrixMulCUDA(float *C, float *A, float *B, int wA, int wB)
{
    
    // Block index
    int bx = blockIdx.x;
    int by = blockIdx.y;

    // Thread index
    int tx = threadIdx.x;
    int ty = threadIdx.y;

    // Index of the first sub-matrix of A processed by the block
    int aBegin = wA * BLOCK_SIZE * by;

    // Index of the last sub-matrix of A processed by the block
    int aEnd   = aBegin + wA - 1;

    // Step size used to iterate through the sub-matrices of A
    int aStep  = BLOCK_SIZE;

    // Index of the first sub-matrix of B processed by the block
    int bBegin = BLOCK_SIZE * bx;

    // Step size used to iterate through the sub-matrices of B
    int bStep  = BLOCK_SIZE * wB;

    // Csub is used to store the element of the block sub-matrix
    // that is computed by the thread
    float Csub = 0;

    // Loop over all the sub-matrices of A and B
    // required to compute the block sub-matrix
    for (int a = aBegin, b = bBegin;
         a <= aEnd;
         a += aStep, b += bStep)
    {

        // Declaration of the shared memory array As used to
        // store the sub-matrix of A
        __shared__ float As[BLOCK_SIZE][BLOCK_SIZE];

        // Declaration of the shared memory array Bs used to
        // store the sub-matrix of B
        __shared__ float Bs[BLOCK_SIZE][BLOCK_SIZE];

        // Load the matrices from device memory
        // to shared memory; each thread loads
        // one element of each matrix
        As[ty][tx] = A[a + wA * ty + tx];
        Bs[ty][tx] = B[b + wB * ty + tx];

        // Synchronize to make sure the matrices are loaded
        __syncthreads();

        // Multiply the two matrices together;
        // each thread computes one element
        // of the block sub-matrix
#pragma unroll

        for (int k = 0; k < BLOCK_SIZE; ++k)
        {
            Csub += As[ty][k] * Bs[k][tx];
        }

        // Synchronize to make sure that the preceding
        // computation is done before loading two new
        // sub-matrices of A and B in the next iteration
        __syncthreads();
    }

    // Write the block sub-matrix to device memory;
    // each thread writes one element
    int c = wB * BLOCK_SIZE * by + BLOCK_SIZE * bx;
    C[c + wB * ty + tx] = Csub;
}

void constantInit(float *data, int size, float val)
{
    for (int i = 0; i < size; ++i)
    {
        //data[i] = val;
        data[i] = (float)drand48();
    }
}

/**
 * Run a simple test of matrix multiplication using CUDA
 */
int matrixMultiply(int argc, char **argv, int block_size, dim3 &dimsA, dim3 &dimsB)
{

	Timer tTotal, tKernel;
	TimerCreate( tTotal );
	TimerCreate( tKernel );

    // Allocate host memory for matrices A and B
    unsigned int size_A = dimsA.x * dimsA.y;
    unsigned int mem_size_A = sizeof(float) * size_A;
    float *h_A = (float *)malloc(mem_size_A);
    unsigned int size_B = dimsB.x * dimsB.y;
    unsigned int mem_size_B = sizeof(float) * size_B;
    float *h_B = (float *)malloc(mem_size_B);

    // Initialize host memory
	srand48( 6834723 );
    const float valB = 0.01f;
    constantInit(h_A, size_A, 1.0f);
    constantInit(h_B, size_B, valB);

    // Allocate host matrix C
    dim3 dimsC(dimsB.x, dimsA.y, 1);
    unsigned int mem_size_C = dimsC.x * dimsC.y * sizeof(float);
    float *h_C = (float *) malloc(mem_size_C);

    if (h_C == NULL)
    {
        fprintf(stderr, "Failed to allocate host matrix C!\n");
        exit(EXIT_FAILURE);
    }

cudaDeviceSynchronize();
TimerStart( tTotal );

    // Allocate device memory
    float *d_A0, *d_B0, *d_C0, *d_A1, *d_B1, *d_C1;


#ifdef CHECK_ERRORS
    cudaError_t error;
    error = cudaMalloc((void **) &d_A0, mem_size_A);

    if (error != cudaSuccess)
    {
        printf("cudaMalloc d_A0 returned error %s (code %d), line(%d)\n", cudaGetErrorString(error), error, __LINE__);
        exit(EXIT_FAILURE);
    }

    error = cudaMalloc((void **) &d_B0, mem_size_B);

    if (error != cudaSuccess)
    {
        printf("cudaMalloc d_B0 returned error %s (code %d), line(%d)\n", cudaGetErrorString(error), error, __LINE__);
        exit(EXIT_FAILURE);
    }

    error = cudaMalloc((void **) &d_C0, mem_size_C);

    if (error != cudaSuccess)
    {
        printf("cudaMalloc d_C0 returned error %s (code %d), line(%d)\n", cudaGetErrorString(error), error, __LINE__);
        exit(EXIT_FAILURE);
    }

    // copy host memory to device
    error = cudaMemcpy(d_A0, h_A, mem_size_A, cudaMemcpyHostToDevice);

    if (error != cudaSuccess)
    {
        printf("cudaMemcpy (d_A0,h_A) returned error %s (code %d), line(%d)\n", cudaGetErrorString(error), error, __LINE__);
        exit(EXIT_FAILURE);
    }
    error = cudaMemcpy(d_B0, h_B, mem_size_B, cudaMemcpyHostToDevice);

    if (error != cudaSuccess)
    {
        printf("cudaMemcpy (d_B0,h_B) returned error %s (code %d), line(%d)\n", cudaGetErrorString(error), error, __LINE__);
        exit(EXIT_FAILURE);
    }
#else

    //Cuda Streams declaration
    cudaStream_t stream0, stream1;
    cudaStreamCreate( &stream0 );
    cudaStreamCreate( &stream1 );
    
    cudaMalloc((void **) &d_A0, mem_size_A);
    cudaMalloc((void **) &d_B0, mem_size_B);
    cudaMalloc((void **) &d_C0, mem_size_C);
    cudaMalloc((void **) &d_A1, mem_size_A);
    cudaMalloc((void **) &d_B1, mem_size_B);
    cudaMalloc((void **) &d_C1, mem_size_C);
    // copy host memory to device
#endif

    // Setup execution parameters
    dim3 threads(block_size, block_size);
    dim3 grid(dimsB.x / threads.x, dimsA.y / threads.y);


    cudaDeviceSynchronize();
TimerStart( tKernel );

    for(int j = 0; j < nMatrix; j++){
        if (block_size == 16)
        {
            if(j%2 == 0){
                cudaMemcpyAsync( d_A0, h_A, mem_size_A, cudaMemcpyHostToDevice, stream0);
                cudaMemcpyAsync( d_B0, h_B, mem_size_B, cudaMemcpyHostToDevice, stream0);
		cudaDeviceSynchronize();
                matrixMulCUDA<16><<< grid, threads, 0, stream1>>>(d_C0, d_A0, d_B0, dimsA.x, dimsB.x);
                cudaMemcpyAsync( h_C, d_C0, mem_size_C, cudaMemcpyDeviceToHost, stream0);
            }else{
                cudaMemcpyAsync( d_A1, h_A, mem_size_A, cudaMemcpyHostToDevice, stream0);
                cudaMemcpyAsync( d_B1, h_B, mem_size_B, cudaMemcpyHostToDevice, stream0);
		cudaDeviceSynchronize();
                matrixMulCUDA<16><<< grid, threads, 0, stream1>>>(d_C1, d_A1, d_B1, dimsA.x, dimsB.x);
                cudaMemcpyAsync( h_C, d_C1, mem_size_C, cudaMemcpyDeviceToHost, stream0);
            }
        } 
        
        else
        {
            if(j%2 == 0){
                cudaMemcpyAsync( d_A0, h_A, mem_size_A, cudaMemcpyHostToDevice, stream0);
                cudaMemcpyAsync( d_B0, h_B, mem_size_B, cudaMemcpyHostToDevice, stream0);
		cudaDeviceSynchronize();
                matrixMulCUDA<32><<< grid, threads, 0, stream1>>>(d_C0, d_A0, d_B0, dimsA.x, dimsB.x);
                cudaMemcpyAsync( h_C, d_C0, mem_size_C, cudaMemcpyDeviceToHost, stream0);
            }else{
                cudaMemcpyAsync( d_A1, h_A, mem_size_A, cudaMemcpyHostToDevice, stream0);
                cudaMemcpyAsync( d_B1, h_B, mem_size_B, cudaMemcpyHostToDevice, stream0);
		cudaDeviceSynchronize();
                matrixMulCUDA<32><<< grid, threads, 0, stream1>>>(d_C1, d_A1, d_B1, dimsA.x, dimsB.x);
                cudaMemcpyAsync( h_C, d_C1, mem_size_C, cudaMemcpyDeviceToHost, stream0);
            }
                    
        }
    }

cudaDeviceSynchronize();
TimerStop( tKernel );

    // Copy result from device to host
#ifdef CHECK_ERRORS
    error = cudaMemcpy(h_C, d_C0, mem_size_C, cudaMemcpyDeviceToHost);

    if (error != cudaSuccess)
    {
        printf("cudaMemcpy (h_C,d_C0) returned error %s (code %d), line(%d)\n", cudaGetErrorString(error), error, __LINE__);
        exit(EXIT_FAILURE);
    }
#else
    cudaMemcpy(h_C, d_C0, mem_size_C, cudaMemcpyDeviceToHost);
#endif

cudaDeviceSynchronize();
TimerStop( tTotal );

printf("SIZE %d\n", dimsA.x );
printf("CLOCK Kernels \n");
printf("%.8lf\t", TimerGetTime(tKernel) );

#ifdef CHECK_ERRORS
    printf("Checking computed result for correctness: ");
    bool correct = true;

    // test relative error by the formula
    //     |<x, y>_cpu - <x,y>_gpu|/<|x|, |y|>  < eps
    double eps = 1.e-6 ; // machine zero

    for (int i = 0; i < (int)(dimsC.x * dimsC.y); i++)
    {
        double abs_err = fabs(h_C[i] - (dimsA.x * valB));
        double dot_length = dimsA.x;
        double abs_val = fabs(h_C[i]);
        double rel_err = abs_err/abs_val/dot_length ;

        if (rel_err > eps)
        {
            printf("Error! Matrix[%05d]=%.8f, ref=%.8f error term is > %E\n", i, h_C[i], dimsA.x*valB, eps);
            correct = false;
        }
    }

    printf("%s\n", correct ? "Result = PASS" : "Result = FAIL");
#endif

    // Clean up memory
    free(h_A);
    free(h_B);
    free(h_C);
    cudaFree(d_A0);
    cudaFree(d_B0);
    cudaFree(d_C0);
    cudaFree(d_A1);
    cudaFree(d_B1);
    cudaFree(d_C1);


    // cudaDeviceReset causes the driver to clean up all state. While
    // not mandatory in normal operation, it is good practice.  It is also
    // needed to ensure correct operation when the application is being
    // profiled. Calling cudaDeviceReset causes all profile data to be
    // flushed before the application exits
    cudaDeviceReset();

#ifdef CHECK_ERRORS
    if (correct)
    {
        return EXIT_SUCCESS;
    }
    else
    {
        return EXIT_FAILURE;
    }
#else
	return EXIT_SUCCESS;
#endif
}


/**
 * Program main
 */
int main(int argc, char **argv)
{

	if ( argc!=2 ) {
		fprintf(stderr, "Usage: %s <matrixSize>\n\n", argv[0]);
		exit(EXIT_FAILURE);
	}
	int SIZE = atoi( argv[1] );
	


    // By default, we use device 0, otherwise we override the device ID based on what is provided at the command line
    int devID = 0;


    cudaError_t error;
    cudaDeviceProp deviceProp;

    error = cudaGetDeviceProperties(&deviceProp, devID);

    if (deviceProp.computeMode == cudaComputeModeProhibited)
    {
        fprintf(stderr, "Error: device is running in <Compute Mode Prohibited>, no threads can use ::cudaSetDevice().\n");
        exit(EXIT_SUCCESS);
    }

    if (error != cudaSuccess)
    {
        printf("cudaGetDeviceProperties returned error %s (code %d), line(%d)\n", cudaGetErrorString(error), error, __LINE__);
    }
    else
    {
        printf("GPU Device %d: \"%s\" with compute capability %d.%d\n\n", devID, deviceProp.name, deviceProp.major, deviceProp.minor);
    }



    // Use a larger block size for Fermi and above
    int block_size = (deviceProp.major < 2) ? 16 : 32;

    dim3 dimsA;
    dim3 dimsB;

    dimsA.x = dimsA.y = dimsB.x = dimsB.y = SIZE;
	
    printf("MatrixA(%d,%d), MatrixB(%d,%d)\n", dimsA.x, dimsA.y, dimsB.x, dimsB.y);

    int matrix_result = matrixMultiply(argc, argv, block_size, dimsA, dimsB);

    exit(matrix_result);
}

